searchNodes=[{"ref":"Error.html","title":"Error","type":"module","doc":""},{"ref":"Maybe.html","title":"Maybe","type":"protocol","doc":""},{"ref":"Maybe.html#is_error?/1","title":"Maybe.is_error?/1","type":"function","doc":"Returns true if the thing is an Error, false otherwise. Examples iex&gt; Maybe.is_error?({:error, nil}) true iex&gt; Maybe.is_error?({:ok, nil}) false"},{"ref":"Maybe.html#is_ok?/1","title":"Maybe.is_ok?/1","type":"function","doc":"Returns true if the thing is not an error, false otherwise. Examples iex&gt; Maybe.is_ok?({:error, nil}) false iex&gt; Maybe.is_ok?({:ok, nil}) true"},{"ref":"Maybe.html#map/2","title":"Maybe.map/2","type":"function","doc":"Applies fun to the value inside an Ok, returning the result lifted into an Ok. Returns an Error untouched if passed one Examples iex&gt; {:ok, 10} |&gt; Maybe.map(fn x -&gt; x * 10 end) {:ok, 100} iex&gt; {:ok, 10} |&gt; Maybe.map(fn x -&gt; {:ok, x * 10} end) {:ok, 100} iex&gt; {:error, 10} |&gt; Maybe.map(fn x -&gt; x * 10 end) {:error, 10} iex&gt; {:error, 10} |&gt; Maybe.map(fn x -&gt; {:error, x * 10} end) {:error, 10} iex&gt; %Error{value: &quot;no&quot;} |&gt; Maybe.map(fn x -&gt; x * 10 end) %Error{value: &quot;no&quot;} iex&gt; %Ok{value: 10} |&gt; Maybe.map(fn x -&gt; x * 10 end) %Ok{value: 100}"},{"ref":"Maybe.html#map_error/2","title":"Maybe.map_error/2","type":"function","doc":"Like map, but works on the error only. If passed an ok value, passes it through untouched. Examples iex&gt; Maybe.map_error({:error, &quot;Nope&quot;}, fn msg -&gt; msg &lt;&gt; &quot; not at all&quot; end) {:error, &quot;Nope not at all&quot;} iex&gt; Maybe.map_error({:ok, &quot;Yes&quot;}, fn msg -&gt; msg &lt;&gt; &quot; not at all&quot; end) {:ok, &quot;Yes&quot;} iex&gt; Maybe.map_error({:error, 10}, fn x -&gt; {:error, x * 10} end) {:error, 100} iex&gt; Maybe.map_error({:error, 10}, fn _ -&gt; {:ok, &quot;Nope&quot;} end) {:ok, &quot;Nope&quot;}"},{"ref":"Maybe.html#unwrap/1","title":"Maybe.unwrap/1","type":"function","doc":"Returns the value contained inside the thing, regardless of whether it is an error or not. Examples iex&gt; Maybe.unwrap({:ok, 10}) 10 iex&gt; Maybe.unwrap({:error, 10}) 10"},{"ref":"Maybe.html#unwrap!/1","title":"Maybe.unwrap!/1","type":"function","doc":"Returns the value contained inside the thing, raising if the thing is an error. Examples iex&gt; Maybe.unwrap!({:ok, 10}) 10 iex&gt; Maybe.unwrap!({:error, &quot;Fluffed it&quot;}) ** (RuntimeError) Error: Fluffed it"},{"ref":"Maybe.html#unwrap_or_else/2","title":"Maybe.unwrap_or_else/2","type":"function","doc":"Returns the value inside the Ok if passed an Ok, or passes the value inside the Error to the or_else fun and calls it. Examples iex&gt; Maybe.unwrap_or_else({:ok, 10}, fn x -&gt; raise x end) 10 iex&gt; Maybe.unwrap_or_else({:error, 10}, fn x -&gt; x + 10 end) 20"},{"ref":"Maybe.html#t:t/0","title":"Maybe.t/0","type":"type","doc":""},{"ref":"Maybe.Pipe.html","title":"Maybe.Pipe","type":"module","doc":"Macro for piping conditionally into the next function. If the value piped in is an error tuple the next function in the pipe will not be called and the value is returned. iex&gt; {:error, 5} ~&gt; Kernel.+(2) If the value piped in is an ok tuple it will be flattened to just the value contained within the tuple when passed to functions in the pipe. The value finally returned from the pipe will be an ok/error tuple. iex&gt; {:ok, 5} ~&gt; Kernel.+(2) iex&gt; {:ok, 5} ~&gt; Kernel.+(2) ~&gt; Kernel.+(3) If the value piped in is not an ok/error tuple a IncorrectMaybePipeError will be raised. In cases where a non-tuple value needs to be passed in a regular pipe (|&gt;) should be used instead. For more information about the motivation for this macro you can start reading about point-free programming. It solves one of the two hardest problems in programming: naming things."},{"ref":"Maybe.Pipe.html#~>/2","title":"Maybe.Pipe.~>/2","type":"macro","doc":""},{"ref":"Ok.html","title":"Ok","type":"module","doc":""},{"ref":"Result.html","title":"Result","type":"protocol","doc":""},{"ref":"Result.html#is_left?/1","title":"Result.is_left?/1","type":"function","doc":""},{"ref":"Result.html#lift/2","title":"Result.lift/2","type":"function","doc":""},{"ref":"Result.html#t:t/0","title":"Result.t/0","type":"type","doc":""},{"ref":"readme.html","title":"Maybex","type":"extras","doc":"Maybex This is a pragmatic implementation of the Maybe monad. It allows you to pipe together functions with abandon, even if they return error values. Why would I use it? Let's look at a completely contrived example. Imagine you get some data and you want to turn it to json then save somewhere: ```elixir {:ok, %{valid?: true, data: \"DATA!\"}} |> turn_into_json() |> save_to_the_db() ``` Let's say they are implemented like this: ```elixir def turn_into_json(%{valid?: false}), do: {:error, \"Nope\"} def turn_into_json(data), do: {:ok, Jason.encode!(data)} def save_to_the_db(json), do: DB.save(json) ``` Notice the problem? The return from `turn_into_json` doesn't match what `save_to_the_db` expects. So we have two options. 1. define `save_to_the_db` such that it can handle an okay / error tuple. 2. use `with`. The first approach would look like this: ```elixir def save_to_the_db({:ok, json}), do: DB.save(json) def save_to_the_db({:error, json}), do: {:error, json} def save_to_the_db(json), do: DB.save(json) ``` There are lots of reasons why it feels wrong. It's not the concern of `save_to_the_db` what `turn_into_json` returns. If `turn_into_json` changes we shouldn't have to also change `save_to_the_db`, so if we do 1. we've introduced coupling that we do not want. Worse than that if we add more functions in between `save_to_the_db` and `turn_into_json` they would also all have to handle an okay / error tuple, which adds overhead. `save_to_the_db` can't handle all of the possible inputs it might get and it shouldn't. In elixir this is easy to do because of pattern matching so is often tempting, but should be avoided. Option 2 looks like this: ```elixir data = {:ok, %{valid?: true, data: \"DATA!\"}} with {:ok, next} {:error, \"Nope\"} end ``` That's much more reasonable, but even this can get unwieldy quickly. If we add more functions, we have to handle them each in the `else` clause, some may return error tuples, some may return nil: ```elixir data = {:ok, %{valid?: true, data: \"DATA!\"}} with {:ok, next} {:error, \"Nope\"} {:error, \"Nope\"} -> {:error, \"Nope\"} end ``` Which again may be fine in small doses, but Maybex offers an alternative: ```elixir {:ok, %{valid?: true, data: \"DATA!\"}} |> Maybe.map(&turn_into_json/1) |> Maybe.map(&save_to_the_db/1) ``` Or even: ```elixir import Maybe.Pipe {:ok, %{valid?: true, data: \"DATA!\"}} ~> turn_into_json() ~> save_to_the_db() ``` How would I use it? Here's how it works. Generally there are two types of things, there are error things and non error things. You can define for yourself what specifically counts as an error, and what isn't, but Maybex provides a few for you. We define the following: | Error | Non Error | | -------------------|------------------| | `{:error, _}` | `{:ok, _}` | | `%Error{value: _}` | `%Ok{value: _}` | If we pass `{:ok, thing}` into [`Maybe.map/2`](Maybe.html#map/2) we will pass `thing` into the mapping function, and return that result wrapped in an okay tuple. If we map over an `{:error, thing}` we wont do anything, and will just return the error tuple: ```elixir iex> {:ok, 10} |> Maybe.map(fn x -> x * 10 end) {:ok, 100} iex> {:error, 10} |> Maybe.map(fn x -> x * 10 end) {:error, 10} iex> {:ok, 10} ...> |> Maybe.map(fn x -> x * 10 end) ...> |> Maybe.map(fn _x -> {:error, \"Nope!\"} end) ...> |> Maybe.map(fn x -> x * 10 end) {:error, \"Nope!\"} iex> %Ok{value: 10} |> Maybe.map(fn x -> x * 10 end) %Ok{value: 100} iex> %Error{value: 10} |> Maybe.map(fn x -> x * 10 end) %Error{value: 10} iex> %Ok{value: 10} ...> |> Maybe.map(fn x -> x * 10 end) ...> |> Maybe.map(fn _x -> %Error{value: \"Nope!\"} end) ...> |> Maybe.map(fn x -> x * 10 end) %Error{value: \"Nope!\"} iex> Maybe.unwrap(%Ok{value: 10}) 10 iex> Maybe.unwrap(%Error{value: 10}) 10 iex> Maybe.map_error(%Error{value: 10}, fn x -> x * 10 end) %Error{value: 100} ``` There is also an infix version of the `map` function which looks like this `~>` ```elixir import Maybe.Pipe iex> {:ok, 10} ~> fn x -> x * 10 end {:ok, 100} iex> {:error, 10} ~> fn x -> x * 10 end {:error, 10} iex> {:ok, 10} ...> ~> fn x -> x * 10 end ...> ~> fn _x -> {:error, \"Nope!\"} end ...> ~> fn x -> x * 10 end {:error, \"Nope!\"} ``` Implementing your own Maybe Type Because Maybex is implemented with protocols you can extend it by implementing Maybe for your own data type. Lets do it for an Ecto.Changeset for no reason whatsoever: ```elixir defmodule Test do use Ecto.Schema embedded_schema do field(:thing, :integer) end end defimpl Maybe, for: Ecto.Changeset do def map(changeset = %{valid?: true}, fun), do: fun.(changeset) def map(changeset, _), do: changeset def map_error(changeset = %{valid?: true}, _), do: changeset def map_error(changeset, fun), do: fun.(changeset) def unwrap!(changeset), do: Ecto.Changeset.apply_action!(changeset, :unwrap) def unwrap(changeset) do with {:ok, ch} ch end end def unwrap_or_else(changeset = %{valid?: true}, _), do: changeset def unwrap_or_else(changeset, fun), do: fun.(changeset) def is_error?(%{valid?: true}), do: false def is_error?(%{valid?: _}), do: true def is_ok?(%{valid?: true}), do: true def is_ok?(%{valid?: _}), do: false end ``` ```sh iex> %Test{} |> Ecto.Changeset.cast(%{thing: \"1\"}, [:thing]) |> Maybe.map_error(fn ch -> Logger.warn(fn -> \"Insert failed: \#{inspect(ch)}\" end) end) #Ecto.Changeset, valid?: true > iex> %Test{} |> Ecto.Changeset.cast(%{thing: false}, [:thing]) |> Maybe.map_error(fn ch -> Logger.warn(fn -> \"Insert failed: \#{inspect(ch)}\" end) end) [warn] Insert failed: #Ecto.Changeset, valid?: false> ``` The Maybe functions The Maybe protocol exposes several functions to help working with optional values. Check the docs but here are some more examples: iex&gt; Maybe.unwrap({:ok, 10}) 10 iex&gt; Maybe.unwrap!({:ok, 10}) 10 iex&gt; Maybe.unwrap({:error, 10}) 10 iex&gt; Maybe.unwrap!({:error, 10}) (RuntimeError) Error: 10 iex&gt; Maybe.map_error({:error, 10}, fn x -&gt; x * 10 end) {:error, 100} iex&gt; Maybe.map_error({:ok, 10}, fn x -&gt; x * 10 end) {:ok, 10} iex&gt; Maybe.unwrap_or_else({:ok, 10}, fn x -&gt; x * 10 end) 10 iex&gt; Maybe.unwrap_or_else({:error, 10}, fn x -&gt; x * 10 end) 100 iex&gt; {:ok, 10} ~&gt; fn x -&gt; x * 10 end |&gt; Maybe.unwrap() 100 There are a list of functions that behave similarly check the docs for more thorough examples."},{"ref":"readme.html#installation","title":"Maybex - Installation","type":"extras","doc":"This is available in Hex, the package can be installed by adding maybex to your list of dependencies in mix.exs: def deps do [ {:maybex, &quot;~&gt; 1.0.0&quot;} ] end Documentation can be found at https://hexdocs.pm/maybex."},{"ref":"readme.html#tests","title":"Maybex - Tests","type":"extras","doc":"To run tests, at the root of the project run mix test"},{"ref":"readme.html#contributing","title":"Maybex - Contributing","type":"extras","doc":"Pull requests and issues are welcome!"}]